VERSION = "2.2.4"

from .paths import templates_path, standard_path, docs_path
from .meta import paralex_factory
from .markdown import to_markdown
from .utils import segment_sounds, create_ids, read_table
from . import strategies





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module generates standard files & documentation.
"""
import json
import re
from itertools import chain
from types import SimpleNamespace

import frictionless as fl
import pandas as pd
from met_brewer import met_brew

# Our modules
from . import VERSION
from .markdown import to_markdown
from .meta import package_from_kwargs, gen_metadata
from .paths import docs_path, standard_path
from .explorer import _gather_dataset_info, _get_glottocode, unknown

import logging

log = logging.getLogger()


def _make_standard_package(*args, **kwargs):
    with (standard_path / "package_spec.json").open('r', encoding="utf-8") as flow:
        package_infos = json.load(flow)
        package_infos["version"] = VERSION

    with (standard_path / "columns_spec.json").open('r', encoding="utf-8") as flow:
        columns = json.load(flow)

    with (standard_path / "files_spec.json").open('r', encoding="utf-8") as flow:
        resources = json.load(flow)["resources"]

    new_resources = []
    for res in resources:
        # replace column names by their full definition
        if res["path"].endswith(".csv"):
            res["schema"]["fields"] = [dict(columns[f]) for f in res["schema"]["fields"]]

        if res["name"] == "forms":
            for col in res["schema"]["fields"]:
                if col["name"] in ["lexeme", "cell"]:
                    col["constraints"] = {"required": True}

        new_resources.append(fl.Resource(res))

    package = package_from_kwargs(resources=new_resources, **package_infos)

    package.to_json(str(standard_path / "paralex.package.json"))


def _build_stats(df, glottocode_data):
    include_path = docs_path / "includes"
    include_path.mkdir(parents=True, exist_ok=True)

    stats = {"summary_pie.md": _summary_family_pie,
             "summary_wanted_fam.md": _summary_wanted_fam,
             "summary_wanted_50.md": _summary_wanted_50,
             "summary_contributors.md": _summary_contributors,
             }

    for name, func in stats.items():
        md = func(df, glottocode_data)
        with (include_path / name).open("w") as f:
            f.write(md)


def _summary_contributors(df, *args, **kwargs):
    def sorter(word):
        word = word.strip()
        m = re.match(r"^.+? [A-Z][^.]", word)
        if not m:
            return word
        return word[m.span()[1] - 2:]

    authors = set(chain(*(c.split(";") for c in df["contributors"])))
    authors - {unknown}
    return "**Dataset contributors: **" + ", ".join(sorted(authors, key=sorter))


def _summary_wanted_50(df, glottolog_data):
    stats = pd.read_csv(docs_path / "language_stats.csv")
    stats_gl = glottolog_data[glottolog_data.Name.isin(stats.Language)].set_index('Name')[
        'Closest_ISO369P3code'].to_dict()
    stats['ISO 693'] = stats.Language.map(stats_gl)
    missing = stats[~stats['ISO 693'].isin(df.lang)].iloc[0:5]
    return missing.set_index("ISO 693").to_markdown()


def _summary_wanted_fam(df, glottolog_data):
    covered = df.family_id.unique()
    lang_labels = dict(glottolog_data.loc[:, ["ID", "Name"]].to_records(index=False))
    fams = glottolog_data.groupby('Family_ID').size().sort_values(ascending=False)
    fams = fams[~fams.index.isin(covered)].iloc[0:5]
    fams.index = fams.index.map(lang_labels)
    fams.index.name = "Language family"
    fams.name = "Number of languages in Glottolog"
    md = fams.to_markdown()
    return md


def _summary_family_pie(df, glottocode_data):
    """Family distribution pie"""
    fams = df.groupby(['color', 'family']).apply(len)
    fams.name = "nb"
    fams = fams.to_frame().reset_index()
    idx = fams.family.to_list()
    values = fams.nb.to_list()
    color = fams.color.to_list()
    mermaid = """
<div style="height: 300px;">
  <canvas id="myChart"></canvas>
</div>

<script src="https://cdn.jsdelivr.net/npm/chart.js"></script>

<script>
  const ctx = document.getElementById('myChart');

  new Chart(ctx, {
    type: 'pie',
    data: {
      labels: """ + str(idx) + """,
      datasets: [{
        backgroundColor: """ + str(color) + """,
        label: 'Number of datasets',
        data: """ + str(values) + """,
      }]
    },
    options: {
        responsive: true,
        maintainAspectRatio: false
    }
  });
</script>

"""
    # for key, value in family_pct.items():
    #     mermaid += f'\n    "{key}": {value}'
    return mermaid


def _build_json(df):
    geojson = {"type": "FeatureCollection", "features": []}

    for _, row in df.dropna(subset=['longitude', 'latitude']).iterrows():
        feature = {"type": "Feature",
                   "geometry": {"type": "Point",
                                "coordinates": [row['longitude'], row['latitude']]
                                }
                   }

        feature['properties'] = {key: row[key] for key in row.index.values[1:]
                                 if key not in ['longitude', 'latitude']}
        geojson['features'].append(feature)

    with (docs_path / 'result.geojson').open('w') as fp:
        json.dump(geojson, fp)


def _build_summaries(glottocode_data):
    """Builds all the summaries from the dataframe of datasets."""
    df = pd.read_csv(docs_path / 'known-datasets.csv')
    # Assign a color to each family
    nitems = df['family_id'].nunique()
    col = met_brew(name="Isfahan2", n=nitems)
    df['color'] = df.groupby(by='family_id').ngroup().apply(lambda x: col[int(x)] if not pd.isna(x) else None)

    _build_json(df)
    _build_stats(df, glottocode_data)


def _write_doc(*args, **kwargs):
    to_markdown(fl.Package(standard_path / "paralex.package.json"),
                docs_path / "specs.md")

    # generate json files for examples
    examples_dir = docs_path / "examples"
    for directory in examples_dir.glob("*/"):
        gen_metadata(SimpleNamespace(config=directory / "paralex-infos.yml", basepath=directory))
    glottocode_data = _get_glottocode()
    _gather_dataset_info(glottocode_data)
    _build_summaries(glottocode_data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Validates packages
"""
import frictionless as fl
from pathlib import Path
import json
import logging
from pybtex.database import parse_file
from .meta import _get_default_package
from .utils import read_table, absolute_path
from . import VERSION
from rich.console import Console
from rich.table import Table
from collections import defaultdict
import re
from enum import IntEnum
from iso639 import Language, LanguageNotFoundError


Status = IntEnum('Status', 'fails passes NA', start=0)

console = Console()
paralex = _get_default_package()


class ParalexValidator(object):
    pass_or_fail = (':X:', ':heavy_check_mark:', '')
    pass_or_flag = (':black_flag:', ':heavy_check_mark:', '')
    must_style = ("bold red", "green", "grey50")
    should_style = ("bold dark_orange3", "green", "grey50")

    def __init__(self, package_path, package_basepath):

        if not package_basepath:
            logging.warning('No basepath passed. Using the parent directory of the package file.')
            self.path = Path(package_path).resolve()
            self.basepath = self.path.parent
        else:
            self.basepath = Path(package_basepath).resolve()
            self.path = self.basepath / package_path

        with open(package_path, "r", encoding="utf-8") as f:
            self.p = fl.Package(json.load(f), basepath=str(self.basepath))

        self.levels = {"musts": [self.is_valid,
                                 self.has_iso_code,
                                 self.has_name,
                                 self.has_doc,
                                 self.has_forms,
                                 self.has_form_triplets,
                                 self.has_expected_foreign_keys,
                                 self.tags_are_defined,
                                 self.if_sources_has_bib,
                                 self.has_vocabulary_relations,
                                 self.has_conventional_cell_ids
                                 ],
                       "checks": [
                           self.has_coherent_analyses,
                           self.is_ipa_or_has_sounds_table,
                           self.has_atomic_forms,
                           self.has_cell_mapping,
                           self.has_feature_values_mapping,
                           self.has_space_separated_phon_form,
                           self.has_sounds_definition,
                           self.all_values_have_cells,
                           self.all_lexemes_have_forms,
                           self.all_cells_have_forms,
                           self.all_columns_have_metadata,
                           self.all_bibkeys_have_ref
                       ],
                       "shoulds": [
                           self.has_recommended_tables,
                           self.has_multiple_cell_mappings,
                           self.has_ids_first,
                           self.has_phon_form,
                           self.has_data_sheet,
                       ]
                       }
        self.styles = {"musts": (self.pass_or_fail, self.must_style),
                       "checks": (self.pass_or_flag, self.should_style),
                       "shoulds": (self.pass_or_fail, self.should_style)
                       }

        self.version = self.p.custom.get('paralex-version', None)
        self.tables = {}
        self.standard_resources = (set(self.p.resource_names) & set(paralex.resource_names))
        self.resource_types = defaultdict(set)

        for name in self.standard_resources:
            self.resource_types[self.p.get_resource(name).type].add(name)

    def check_version(self):
        """Issue a warning if the standard metadata does not correspond to the version of the validator."""
        if not self.version:
            logging.warning(f"Your package doesn't contain any version number. It was probably generated with version 1.0 of the standard.\nVersion 1.0 is now deprecated and you should upgrade at least to version 2.0. Current validator is for {VERSION}.")
        elif self.version != VERSION:
            logging.warning(f'You are using the version {VERSION} of the validator, but your metadata was generated for version {self.version}.\nIn case you encounter validation issues, please consider upgrading your package or downgrading the validator.')

    def get_table(self, name):
        if name not in self.tables:
            self.tables[name] = read_table(name, self.p)
        return self.tables[name]

    def has_name(self):
        if self.p.name is not None:
            return (Status.passes, "Dataset has a name", "")
        return (Status.fails, "Package is missing a name attribute", "")

    def is_valid(self):
        report = fl.validate(self.path)
        if report.valid:
            return (report.valid, "Pass frictionless validation", "")
        return (report.valid,
                "Did not pass frictionless validation.",
                "Run `frictionless validate <package>` for details.")

    def has_iso_code(self):
        if "languages_iso639" not in self.p.custom:
            return (False, "Package is missing a languages_iso639 list. This is mandatory from Paralex v.2.0.0", "Add a list of 3-letter language codes")
        codes =  self.p.custom["languages_iso639"]
        if not codes:
            return (False, "Package has an empty languages_iso639 list. This is mandatory from Paralex v.2.0.0", "Add a list of 3-letter language codes")
        errors = []
        for code in codes:
            try:
                Language.match(code)
            except LanguageNotFoundError:
                errors.append(code)
        if errors:
            return (False,
                    f"The following are not valid ISO 639 codes: {', '.join(errors)}",
                    "Use only valid ISO 639 codes")
        return (True, "The languages_iso639 list is present and valid", "")


    def validate(self, level):
        icons, styles = self.styles[level]
        total = 0
        passed = 0
        for task in self.levels[level]:
            try:
                status, report, solution = task()
            except Exception as e:
                status, report, solution = False, str(e), ""
            passed += status == 1
            total += status < 2
            console.print("- " + icons[status] + ' ' + report, style=styles[status])
            if solution:
                console.print(" " + solution, style=styles[status])
        return passed, total

    def has_expected_foreign_keys(self):

        error = "Could not find some foreign keys:\n\t- {}"
        suggestion = "Re-generate metadata using the paralex package"
        invalid_keys = []
        for table_name in self.resource_types['table']:
            table = self.p.get_resource(table_name)
            table_def = paralex.get_resource(table_name)

            table_refs = {
                (tuple(rel["fields"]),
                 rel["reference"]["resource"],
                 tuple(rel["reference"]["fields"]),
                 ) for rel in table.schema.foreign_keys}

            for rel in table_def.schema.foreign_keys:
                cols = tuple(rel["fields"])
                foreign_table = rel["reference"]["resource"]
                foreign_cols = tuple(rel["reference"]["fields"])
                ref = (cols, foreign_table, foreign_cols)

                if foreign_table in self.resource_types['table'] and ref not in table_refs:
                    invalid = f"{table_name}.{'+'.join(cols)} => {foreign_table}.{'+'.join(foreign_cols)}"
                    invalid_keys.append(invalid)

        if invalid_keys:
            return (Status.fails, error.format("\n\t- ".join(invalid_keys)), suggestion)
        return (Status.passes, "Has all expected relations", "")

    def all_lexemes_have_forms(self):
        error = "Not all lexemes have forms: {}"
        if "lexemes" in self.standard_resources:
            lex = set(self.get_table("lexemes").lexeme_id.to_list())
            forms = set(self.get_table("forms").lexeme.to_list())
            unref_lex = lex - forms
            if unref_lex:
                return (Status.fails, error.format(", ".join(map(str, unref_lex))),
                        "Check whether there are missing forms.")
            return (Status.passes, "All lexemes have forms.", "")
        return (Status.NA, "(No lexemes table)", "")

    def all_cells_have_forms(self):
        error = "Not all cells have forms: {}"
        if "cells" in self.standard_resources:
            cells = set(self.get_table("cells").cell_id.to_list())
            forms = set(self.get_table("forms").cell.to_list())
            unref_cells = cells - forms
            if unref_cells:
                return (Status.fails, error.format(", ".join(unref_cells)),
                        "Check whether there are missing forms.")
            return (Status.passes, "All cells have forms.", "")
        return (Status.NA, "(No cells table)", "")

    def all_values_have_cells(self):
        error = "Not all features have cells: {}"
        if "features-values" in self.standard_resources and "cells" in self.standard_resources:
            cells = set((part
                         for cell in self.get_table("cells").cell_id.to_list()
                         for part in cell.split(".")
                         ))
            vals = set(self.get_table("features-values").value_id.to_list())
            unref_vals = vals - cells
            if unref_vals:
                return (Status.fails, error.format(", ".join(unref_vals)),
                        "Check whether there are missing cells.")
            return (Status.passes, "All feature values are present in some cells.", "")
        return (Status.NA, "(No foreign cell<->features relations)", "")

    def all_bibkeys_have_ref(self):
        error = "Not all citation keys have references: {}"
        keys = set()
        for table in self.resource_types['table']:
            df = self.get_table(table)
            if 'source' in df.columns:
                keys.update(set(df.source.unique()))
        if keys:
            if "sources" in self.standard_resources:
                bib = parse_file(absolute_path(self.p, self.p.get_resource('sources').path))
                missing = keys - set(bib.entries.keys())
                if missing:
                    return (Status.fails, error.format(", ".join(missing)),
                            "Check whether the .bib file is complete.")
                return (Status.passes, "All citation keys correctly link to the .bib file.", "")
            return (Status.passes, "No .bib file could be found for the citation keys.", "")
        return (Status.NA, "(No citation keys)", "")

    def all_columns_have_metadata(self):
        no_metadata = []
        for r in self.p.resources:
            for col in r.schema.fields:
                if not col.description:
                    no_metadata.append(r.name + "." + col.name)
        if no_metadata:
            return (Status.fails,
                    "These columns have no detailed metadata: " + ", ".join(no_metadata),
                    "Edit the gen-metadata script to add type, description, title, and constraints.")
        return (Status.passes, "All columns have rich metadata", "")

    def tags_are_defined(self):
        taggable_tables = {"forms", "lexeme", "frequencies"} & self.standard_resources

        by_tag = defaultdict(list)
        if "tags" in self.standard_resources:
            for r in self.p.get_resource("tags").read_rows():
                by_tag[r["tag_column_name"]].append(r["tag_id"])

        tags_used = False
        for tab_name in taggable_tables:
            table = self.p.get_resource(tab_name)
            tag_cols = {col for col in table.schema.field_names
                        if col.endswith("_tag")}
            if tag_cols:
                tags_used = True
                if "tags" not in self.standard_resources:
                    return (Status.fails, "Tags table missing", "Add a tags table")
        if tags_used:
            return (Status.passes, "Tags table is present", "")
        return (Status.NA, "(No tags used)", "")

    def has_vocabulary_relations(self):
        forms_table = self.p.get_resource("forms")
        missing_rels = []

        rels = [("graphemes", "grapheme_id", "forms", "orth_form"),
                ("sounds", "sound_id", "forms", "phon_form"),
                ("features-values", "value_id", "cells", "cell_id"),
                ]

        for foreign_table, identifier, table, col in rels:
            if {foreign_table, table} <= self.standard_resources \
                    and forms_table.schema.has_field(col):
                form = forms_table.schema.get_field(col)
                if not form.constraints:
                    missing_rels.append(f"{foreign_table}.{identifier} ~~~> {table}.{col}")

        if missing_rels:
            suggestion = "Re-generate metadata using the paralex package"
            return (Status.fails, "Vocabulary relation(s) missing:".format("\n\t- ".join(missing_rels)), suggestion)
        return (Status.passes, "Has vocabulary constraints", "")


    def has_doc(self):
        if "readme" in self.resource_types["text"]:
            return (Status.passes, "Has a readme file", "")
        doc_file = list(Path(self.path).parent.glob("[Rr][Ee][Aa][Dd][Mm][Ee]*"))
        if not doc_file:
            return (Status.fails, "Could not find any readme file", "Add a readme.md")
        return (Status.NA,
                f"(deprecation) readme file ({str(doc_file[0].relative_to(Path(self.p.basepath).resolve()))}) is not listed in the metadata",
                "Add it to the metadata.")

    def has_data_sheet(self):
        if "data_sheet" in self.resource_types["text"]:
            return (Status.passes, "Has a data sheet file", "")
        else:
            data_sheet_re = "*[Dd][Aa][Tt][Aa]_[Ss][Hh][Ee][Ee][Tt]*"
            doc_file = list(Path(self.path).parent.glob(data_sheet_re))
            if not doc_file:
                doc_file = list(Path(self.path).parent.glob("docs/" + data_sheet_re))
            if not doc_file:
                return (Status.fails, "Could not find any data sheet file", "Add a data sheet!")
            return (Status.fails, f"(deprecation) Data sheet ({str(doc_file[0].relative_to(Path(self.p.basepath).resolve()))}) is not listed in the metadata", "Add it to the metadata.")

    def if_sources_has_bib(self):
        def has_sources():
            for r in self.resource_types["table"]:
                t = self.p.get_resource(r)
                if "source" in t.schema.field_names:
                    return True
            return False
        if has_sources():
            if "sources" not in self.resource_types["file"]:
                return (Status.fails, "Has `source` column(s), but no declared source (bib) file.", "Add a bib file.")
            return (Status.passes, f"Has source columns and a source file", "")
        return (Status.NA, f"(No sources columns)", "")


    def has_forms(self):
        exists = "forms" in set(self.p.resource_names)
        if not exists:
            return (Status.fails, "Missing forms table !", "Add a forms table")
        return (Status.passes, "Has a forms table", "")


    def has_form_triplets(self):
        forms = self.p.get_resource("forms")
        cols = set(forms.schema.field_names)
        expected = {"form_id", "lexeme", "cell"}
        one_of = {"phon_form", "orth_form"}
        if not expected <= cols and not one_of & cols:
            return (Status.fails, "Forms need to minimally define columns "
                                  "form_id, lexeme, cell, as well as either "
                                  "phon_form or orth_form.", "Add missing columns")
        return (Status.passes, "Forms table minimally defines triplets", "")


    def has_space_separated_phon_form(self):
        forms_table = self.get_table("forms")
        cols = forms_table.columns
        if "phon_form" not in cols:
            return (Status.NA, "(No phonological forms)", "")
        has_space = (forms_table["phon_form"].str.contains(" ", regex=False)
                     | (forms_table["phon_form"].str.len() <= 5))
        if not has_space.all():
            no_spaces = forms_table.loc[~has_space, :].head(5).to_markdown()
            return (Status.fails, "Phonological forms seem to not be space-separated.\n" + no_spaces,
                    "Ensure phonemes are separated by spaces.")

        return (Status.passes, "Phonological forms, if present, are space separated.", "")


    def has_coherent_analyses(self):
        forms_table = self.get_table("forms")
        cols = forms_table.columns
        if "analysed_phon_form" not in cols and "analysed_orth_form" not in cols:
            return (Status.NA, f'(No analysis columns).', '')
        analysis_rels = [("phon_form", "analysed_phon_form"),
                         ("orth_form", "analysed_orth_form"),
                         ]
        if "phon_form" in forms_table.columns:
            forms_table = forms_table[forms_table.phon_form != "#DEF#"]
        else:
            forms_table = forms_table[forms_table.orth_form != "#DEF#"]
        for form, analysis in analysis_rels:
            if form in cols and analysis in cols:
                coherent = forms_table[form].str.replace(" ", "") == \
                           forms_table[analysis].str.replace(r"[+#\s]", "", regex=True)
                if not coherent.all():
                    ex = str(forms_table[~coherent].iloc[:3, :].to_markdown())
                    return (Status.fails, f"Analyses do not seem to correspond to forms, for example: \n{ex}\n "
                                          f"Or perhaps you used custom separators ?",
                            "Make sure the forms are identical except for separators. Define separators in the docs if custom.")
        return (Status.passes, "Analyses are coherent with forms", "")


    def has_conventional_cell_ids(self):
        forms_table = self.get_table("forms")
        incorrect = []
        for c in forms_table.cell.unique():
            if not re.match(r"[\da-z]([\da-z\.]*[\da-z]+)", c):
                incorrect.append(c)
        if incorrect:
                return (Status.fails, f"Cell format is incorrect in: {', '.join(incorrect)}.",
                                    "Make sure the cells are lowercase sequences of feature values, separated by dots.")
        return (Status.passes, "Cell format is correct", "")


    def is_ipa_or_has_sounds_table(self):
        if "sounds" in self.p.resource_names:
            return (Status.passes, "Defines sounds explicitly", "")

        forms = self.get_table("forms")
        cols = forms.columns
        if "phon_form" not in cols:
            return (Status.NA, "Sounds file is not necessary, there are no phon_form", "")

        ipa = ("['ɶʞ̄lɓ̃ʔç̥̜|ǁ̪̀.ʃ˞̙ɣ̟↗j̤ʙχβtɞɲz̼’eʈɤɕð̢̹ʉɦ̈ꜛʰ̰yʌˑǂɠʵɥ͡ʏɾs‿ʒ"
               "ǃʊ ̴̠̩̫̂ɝħʳʍ̊ʋʝøɽ̮œʲfæhʛrˈ˔ɗ́ŋˤɐbʁqʶˀcʼɚɯ\u2009ɨʜθx̻ʀɮ"
               "ɜo᷇ɧ̞ɢː᷅ɪ̘ɭʘɸvʐʎʴɹɘəɡdⱳʕ̽ɛ̺ʡʱ↘wˠ̆ɱ˚ɟˡɵʂꜜɳʑk̬͜ˌ̋᷄ǔʄɺʢɬⱱɖɰȁ"
               "ǀ᷆̚ɑ᷉ʷɴiʟ̯pɒmᶑɔɻ̝᷈n' ]+")

        matches = forms.phon_form.str.fullmatch(ipa)
        if not matches.all():
            return (Status.fails, "Phonological notation seems to have some custom elements, but no sounds file found.",
                    "Add a sounds file to define phonemes")
        return (Status.passes, "No sounds file, but the phon_form notation looks like IPA.", "")


    def has_atomic_forms(self):
        forms = self.get_table("forms")
        cols = forms.columns
        separators = r".*[;/|].*"

        if "phon_form" in cols:
            has_seps = forms.phon_form.str.fullmatch(separators).any()
            if has_seps:
                return (Status.fails, "Some phon_form values  seem to contain separators !",
                        "Use separate rows for overabundance (long form).")
        if "orth_form" in cols:
            has_seps = forms.orth_form.str.fullmatch(separators).any()
            if has_seps:
                return (Status.fails, "Some orth_form values seem to contain separators !",
                        "Use separate rows for overabundance (long form).")

        return (Status.passes, "Inflected forms seem atomic (no separators found).", "")


    def has_cell_mapping(self):
        if not "cells" in self.p.resource_names:
            return (Status.NA, "(No cells table)", "")
        cells = self.p.get_resource("cells")
        cols = set(cells.schema.field_names)
        perhaps_mappings = cols - {'cell_id', 'label', 'comment', 'POS', 'frequency', 'canonical_order'}
        if not perhaps_mappings:
            return (Status.fails, "It seems that cells are not mapped to any other vocabulary",
                    "Add mappings (eg. to UD, unimorph, other corpora schemes) to improve interoperability")
        return (Status.passes,
                "It seems that cells are mapped to other vocabularies.", "")


    def has_feature_values_mapping(self):
        if not "features-values" in self.p.resource_names:
            return (Status.NA, "(No features-values table)", "")
        values = self.p.get_resource("features-values")
        cols = set(values.schema.field_names)
        perhaps_mappings = cols - {'value_id', 'label', 'feature', 'comment', 'POS', 'canonical_order'}
        if not perhaps_mappings:
            return (Status.fails, "It seems that features are not mapped to any other vocabulary",
                    "Add mappings (eg. to UD, unimorph, other corpora schemes) to improve interoperability")
        return (Status.passes,
                "It seems that features are mapped to other vocabularies.", "")


    def has_sounds_definition(self):
        if not "sounds" in self.p.resource_names:
            return (Status.NA, "(No sounds table)", "")
        values = self.p.get_resource("sounds")
        cols = set(values.schema.field_names)
        perhaps_defs = cols - {'sound_id', 'comment'}
        if not perhaps_defs:
            return (Status.fails, "It seems that sounds are missing explicit labels or definitions.",
                    "Add a label column or map to existing vocabularies to improve interoperability")
        return (Status.passes,
                "It seems that sounds have definitions or labels.", "")


    def has_multiple_cell_mappings(self):
        if not "cells" in self.p.resource_names:
            return (Status.NA, "(No cells table)", "")
        cells = self.p.get_resource("cells")
        cols = set(cells.schema.field_names)
        perhaps_mappings = cols - {'cell_id', 'label', 'comment', 'POS', 'frequency', 'canonical_order'}
        if len(perhaps_mappings) < 2:
            return (Status.fails, "It would be better to map cells to multiple vocabularies",
                    "Add more mappings (eg. to UD, unimorph, other corpora schemes) to improve interoperability")
        return (Status.passes,
                "It seems that cells are mapped to multiple vocabularies.", "")


    def has_recommended_tables(self):
        missing = {"cells", "features-values"} - self.resource_types['table']
        forms = self.p.get_resource("forms")
        cols = set(forms.schema.field_names)
        if "phon_form" in cols and not "sounds" in self.standard_resources:
            missing.add("sounds")

        if missing:
            return (Status.fails, "Dataset is missing table(s): " + ", ".join(missing),
                    "Add any missing table")
        return (Status.passes, "Dataset has all recommended tables", "")


    def has_ids_first(self):
        resources = set(self.p.resource_names)
        failing = []
        for r in resources:
            res = self.p.get_resource(r)
            if res.tabular:
                cols = res.schema.field_names
                if not cols[0] in res.schema.primary_key:
                    failing.append(r)
        if failing:
            return (Status.fails, "Tables where the first column isn't the id: " + ", ".join(failing),
                    "Make sure ids are present, and the first column")
        return (Status.passes, "Dataset has ids in the right place", "")


    def has_phon_form(self):
        forms = self.p.get_resource("forms")
        cols = set(forms.schema.field_names)
        if not "phon_form" in cols:
            return (Status.fails, "Forms table does not have phon_form column",
                    "Adding phonological transcriptions would make the dataset more useful for linguistic investigation.")
        return (Status.passes, "Form table has phon_form column", "")


def paralex_validation(args):

    p = ParalexValidator(args.package, args.basepath)
    p.check_version()

    console.rule(f"[bold]Paralex v.{VERSION} validation for {Path(args.package).resolve().relative_to(p.basepath)}")

    console.print("\nChecking MUSTs...", style="bold")
    console.print(
        "These are mandatory: any failing statement needs to be corrected for the dataset to comply with paralex.")
    m = p.validate("musts")
    console.print("\nLooking for potential issues...", style="bold")
    console.print("These are indicative: failing statement indicate potential mistakes, to be checked.")
    c = p.validate("checks")

    console.print("\nChecking SHOULDs...", style="bold")
    console.print("These are recommendations: any failing statement indicates a potential improvement of the dataset.")
    s = p.validate("shoulds")

    return {"must": m, "checks":c, "shoulds":s}


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module provides paths to various subfolders
"""
from pathlib import Path

standard_path = (Path(__file__).parent / "standard/").resolve()
templates_path = (Path(__file__).parent / "templates/").resolve()
docs_path = (Path(__file__).parent / "docs/").resolve()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module provides templates to render paralex metadata as markdown files.
"""
from pathlib import Path
import pkg_resources
import shutil
from .paths import templates_path


class CustomTemplates(object):
    def __init__(self, *files):
        self.replace = [(templates_path / f,
                         pkg_resources.resource_filename("frictionless",
                                                         f"assets/templates/{f}"),
                         pkg_resources.resource_filename("frictionless",
                                                         f"assets/templates/{f}.tmp"))
                        for f in files]

    def __enter__(self):
        for custom, prev, tmp in self.replace:
            shutil.copyfile(prev, tmp)
            shutil.copyfile(custom, prev)

    def __exit__(self, type, value, traceback):
        for custom, prev, tmp in self.replace:
            shutil.copyfile(tmp, prev)
            Path(tmp).unlink()


def to_markdown(package, output_filename, title=None):
    yaml_lines = ["warning: This file was automatically generated, do NOT EDIT"]
    if title is not None:
        yaml_lines +=  ["title: "+title]
    yaml_lines= "\n    " + "\n    ".join(yaml_lines)

    with CustomTemplates("field.md", "package.md", "resource.md"):
        with open(output_filename, "w", encoding="utf-8") as f:
            f.write("---"+yaml_lines+"\n---\n\n\n")
            f.write(package.to_markdown())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
from hypothesis import strategies as st

Cs = {'ʣ': 'ds', 'ʦ': 'ts', 'ɹ': 'r', 'l': 'l', 'ɫ': 'l', 'ɮ': 'lz', 'z': 'z', 'ɬ': 'l', 's': 's', 'n': 'n', 'd': 'd',
      't': 't', 'ɺ': 'r', 'ɾ': 'r', 'r': 'r',
      'ȶ': 't', 'ȡ': 'd', 'ʥ': 'dz', 'ʨ': 'č', 'ȴ': 'll', 'ʑ': 'ź', 'ɕ': 'š', 'ȵ': 'ny', 'ɸ': 'f', 'ɓ': 'b', 'm': 'm',
      'b': 'b', 'p': 'p', 'ʙ': 'b', 'ð': 'dh', 'ɗ': 'ɗ', 'ʢ': 'ʢ', 'ʜ': 'x', 'ʡ': 'g', 'ɦ': 'h', 'h': 'h', 'ʔ': '\'',
      'ɥ': 'w', 'w': 'w', 'ʍ': 'hw', 'ʋ': 'w', 'v': 'v', 'f': 'f', 'ɱ': 'm', 'j': 'y', 'ʎ': 'ly', 'ʝ': 'gl',
      'ç': 'j', 'ɟ': 'gu', 'c': 'tj', 'ʕ': 'r', 'ʆ': 'shl', 'ʤ': 'dj', 'ʧ': 'tsh', 'ʒ': 'j',
      'ʃ': 'sh', 'ɻ': 'r', 'ɭ': 'l', 'ʐ': 'zh', 'ʂ': 'sh', 'ɳ': 'rn', 'ɖ': 'd', 'ʈ': 't', 'ɽ': 'r', 'ʁ': 'r', 'ɴ': 'n',
      'ɢ': 'g', 'q': 'q', 'ʀ': 'r', 'ɰ': 'g', 'ʟ': 'l', 'ɣ': 'x', 'x': 'x', 'ɠ': 'g', 'ŋ': 'ng', 'g': 'g', 'k': 'k'}

Vs = {'a': 'a', 'ã': 'an', 'e': 'é', 'i': 'i', 'ĩ': 'in', 'o': 'o', 'õ': 'on', 'u': 'oo',
      'ũ': 'un', 'y': 'u', 'æ': 'ae', 'ø': 'eu', 'œ': 'oe', 'ɐ': 'e', 'ɑ': 'a', 'ɒ': 'a', 'ɔ': 'o', 'ɘ': 'e', 'ə': '',
      'ɛ': 'e', 'ɜ': 'e', 'ɞ': 'e', 'ɤ': 'u', 'ɨ': 'i', 'ɪ': 'u', 'ɯ': 'oo', 'ɵ': 'ö', 'ɶ': 'ø',
      'ʉ': 'uu', 'ʊ': 'o', 'ʌ': 'o', 'ʏ': 'ü'}

abbr = st.text("bcdfghjklmnpqrstvwxyzaeiou", min_size=1, max_size=3)

POS = [
    "verb",
    "numeral",
    "conjunction",
    "noun",
    "adposition",
    "determiner",
    "article",
    "adverb",
    "pronoun",
    "fusedPreposition",
    "adjective",
    "symbol",
    "particle",
    "conditionalParticle",
    "demonstrativePronoun",
    "interjection",
    "semiColon",
    "diminutiveNoun",
    "possessivePronoun",
    "prepositionalAdverb",
    "compoundPreposition",
    "interrogativeRelativePronoun",
    "possessiveParticle",
    "plainVerb",
    "letter",
    "interrogativeDeterminer",
    "relativePronoun",
    "postposition",
    "fusedPronounAuxiliary",
    "interrogativeOrdinalNumeral",
    "indefiniteOrdinalNumeral",
    "strongPersonalPronoun",
    "possessiveRelativePronoun",
    "ordinalAdjective",
    "collectivePronoun",
    "commonNoun",
    "infinitiveParticle",
    "comparativeParticle",
    "partitiveArticle",
    "invertedComma",
    "lightVerb",
    "emphaticPronoun",
    "distinctiveParticle",
    "genericNumeral",
    "possessiveAdjective",
    "reflexivePossessivePronoun",
    "colon",
    "coordinationParticle",
    "presentParticipleAdjective",
    "fusedPrepositionPronoun",
    "cardinalNumeral",
    "indefiniteDeterminer",
    "numeralFraction",
    "questionMark",
    "generalAdverb",
    "superlativeParticle",
    "point",
    "indefiniteMultiplicativeNumeral",
    "comma",
    "closeParenthesis",
    "futureParticle",
    "personalPronoun",
    "reflexivePersonalPronoun",
    "adverbialPronoun",
    "reciprocalPronoun",
    "openParenthesis",
    "pastParticipleAdjective",
    "negativePronoun",
    "relativeDeterminer",
    "existentialPronoun",
    "pronominalAdverb",
    "relativeParticle",
    "exclamativeDeterminer",
    "multiplicativeNumeral",
    "reflexiveDeterminer",
    "modal",
    "unclassifiedParticle",
    "properNoun",
    "allusivePronoun",
    "interrogativeCardinalNumeral",
    "bullet",
    "subordinatingConjunction",
    "irreflexivePersonalPronoun",
    "possessiveDeterminer",
    "negativeParticle",
    "indefinitePronoun",
    "generalizationWord",
    "coordinatingConjunction",
    "deficientVerb",
    "adjective-i",
    "impersonalPronoun",
    "indefiniteCardinalNumeral",
    "adjective-na",
    "qualifierAdjective",
    "affirmativeParticle",
    "mainVerb",
    "fusedPrepositionDeterminer",
    "indefiniteArticle",
    "weakPersonalPronoun",
    "suspensionPoints",
    "interrogativeMultiplicativeNumeral",
    "affixedPersonalPronoun",
    "auxiliary",
    "circumposition",
    "copula",
    "demonstrativeDeterminer",
    "participleAdjective",
    "exclamativePoint",
    "interrogativePronoun",
    "presentativePronoun",
    "punctuation",
    "definiteArticle",
    "slash",
    "exclamativePronoun",
    "preposition",
    "conditionalPronoun",
    "relationNoun",
    "interrogativeParticle"
]


@st.composite
def tags(draw, suffix):
    """ Strategy to generate fake tag values (for overabundance_tag, etc) """
    return draw(st.text(min_size=3, max_size=6)) + suffix


lexeme_cols = {
    "inflection_class": st.none() | st.text(min_size=1, max_size=6),
    "source": st.none() | st.text(min_size=4, max_size=10),
    "frequency": st.none() | st.floats(),
    "label": None,
    "gloss": st.none() | st.text(min_size=5, max_size=10),
    "meaning": st.none() | st.text(min_size=5, max_size=50),
    "comment": st.none() | st.text(),
    "POS": st.sampled_from(POS),
    "language_ID": st.text(min_size=3, max_size=6),
    "analysis_tag": tags("_analysis_tag"),
    "defectiveness_tag": tags("_def_tag"),
    "overabundant_tag": tags("_ov_tag"),
    "epistemic_tag": tags("_epis_tag"),
    "variants_tag": tags("_variant_tag")
}


@st.composite
def sound_inventory(draw):
    """ Strategy to generate a sound inventory as a dict """
    C = draw(st.sets(st.sampled_from(list(Cs)), min_size=10, max_size=len(Cs) // 2))
    V = draw(st.sets(st.sampled_from(list(Vs)), min_size=3, max_size=10))
    return {"C": list(C), "V": list(V)}


@st.composite
def word(draw, sound_inventory, min_size=2, max_size=None):
    """ Strategy to generate a word """
    C = st.sampled_from(sound_inventory["C"])
    V = st.sampled_from(sound_inventory["V"])
    size = draw(st.integers(min_value=min_size, max_value=max_size or 10))
    r = draw(st.randoms())
    cats = (V, C) if int(r.random() < 0.5) else (C, V)
    phonemes = [draw(cats[0]) if i % 2 == 0 else draw(cats[1]) for i in range(size)]
    letters = [Cs.get(x, Vs.get(x)) for x in phonemes]
    return " ".join(phonemes), "".join(letters)


@st.composite
def word_with_sound_inventory(draw, **kwargs):
    """ Strategy to generate a word, without requiring an existing sound inventory. """
    return draw(word(draw(sound_inventory()), **kwargs))


@st.composite
def feature_structures(draw):
    """ Strategy to create synthetic feature structures

    Featurestructures look like:
        {   feat: [val1, val2],
        }

    We want all abbreviations, be them features or values, to be unique.

    Args:
        draw:

    Returns:

    """
    # Pick a vocabulary of abbreviations (a unique set of abbreviations)
    abbrs = draw(st.sets(abbr, min_size=6, max_size=300))

    # Select a set of abbreviations to serve as features
    #  and remove them from the vocabulary
    features = draw(st.sets(st.sampled_from(sorted(abbrs)),
                            min_size=1,
                            max_size=len(abbrs) // 2))
    abbrs = abbrs - features
    # Build the dictionnary of feature => { frozenset({value}), ... }
    # Sample some abbreviations to serve as values for each feature,
    # removing them from the vocabulary each time
    # until either features or vocabulary are exhausted
    fs = {}
    for f in features:
        values = draw(st.sets(st.sampled_from(sorted(abbrs)), min_size=2, max_size=20))
        abbrs = abbrs - values
        fs[f.upper()] = values
        if len(abbrs) < 2:
            break
    return fs


@st.composite
def cell(draw, fs, features):
    """ Strategy to create synthetic cells

    cells are frozensets of values,
     such that each value is taken from a distinct feature.

    Args:
        draw:
        fs: a feature_structure

    Returns:

    """
    # pick one value in each feature
    values = [draw(st.sampled_from(sorted(fs[f]))) for f in features]
    return ".".join(values)


@st.composite
def features_conjunctions(draw, fs):
    """ Strategy to create synthetic feature conjunctions"""
    fs = sorted(fs)
    # Pick how many features in each set
    sizes = draw(st.lists(st.integers(1, len(fs)), min_size=1))
    f_tuples = []
    # Iterate over feature set sizes
    for s in sorted(sizes, reverse=True):
        # Pick a set of features (eg: number & case)
        f = draw(st.sets(st.sampled_from(fs), min_size=s, max_size=s))
        # Check we don't already have a superset (eg: number & case & gender)
        is_included = any((f <= f2 for f2 in f_tuples))
        if not is_included:
            f_tuples.append(f)
    return f_tuples


@st.composite
def cell_feats(draw, min_size=2, max_size=50):
    """ Strategy to create synthetic cells and features"""
    features = draw(feature_structures())
    feature_conjs = draw(features_conjunctions(features))
    cells = set()
    for feats in feature_conjs:
        new_cells = draw(st.sets(cell(features, feats), min_size=min_size, max_size=max_size))
        cells |= new_cells
        max_size -= len(new_cells)
        if max_size <= min_size:
            break
    return list(cells), features


@st.composite
def orth_word(draw, inventory, **kwargs):
    """ Strategy to create synthetic orthographic word"""
    phon, orth = draw(word(inventory, **kwargs))
    return orth


@st.composite
def phon_word(draw, inventory, **kwargs):
    """ Strategy to create synthetic phonological word"""
    phon, orth = draw(word(inventory, **kwargs))
    return phon


@st.composite
def affix(draw, inventory):
    """ Strategy to create synthetic affix"""
    r = draw(st.randoms())
    form = draw(word(inventory, min_size=0, max_size=3))
    is_suffix = r.random() < 0.5
    return (is_suffix, form)

@st.composite
def lexemes_list(draw, sounds, lexemes_size=None):
    word_strategy = word(sounds, min_size=3, max_size=6)
    citation_forms = st.sets(word_strategy,
                             min_size=lexemes_size or 2,
                             max_size=lexemes_size or 10)
    return sorted(draw(citation_forms))


@st.composite
def lexemes_table(draw, lexemes, sounds):
    df = pd.DataFrame([(b, a) for (a, b) in lexemes],
                      columns=["lexeme_id", "label"])
    extra_cols = draw(st.lists(st.sampled_from(lexeme_cols), unique=True))
    for col in extra_cols:
        if col == "label":
            df[col] = lexemes["lexeme_id"]
            df["lexeme_id"] = ["lexeme_n_"+str(i+1) for i in range(df.shape[0])]
        else:
            df[col] = [draw(lexeme_cols[col]) for _ in range(df.shape[0])]
    df = df.set_index("lexeme_id")
    return df

@st.composite
def paralex_dataset(draw, lexemes_size=None, cells_size=None):
    """ This represents a full paralex dataset

    Returns:
        a strategy to generate false long form paradigms
    """
    # Mandatory: forms table
    # Mandatory: metadata df
    # Optional: feature-values; cells; lexemes; sounds; graphemes

    m = draw(metadata())
    sounds = draw(sound_inventory())
    lexemes = draw(lexemes_list(sounds, lexemes_size=lexemes_size))
    cells, feats = draw(cell_feats(min_size=cells_size or 2, max_size=cells_size or 50))
    form_table = draw(forms_df(sounds, lexemes, cells))
    files = {"forms.csv": form_table}
    m["resources"].append({"path": "forms.csv",
                           "name": "forms"})
    extra_tables = draw(st.lists(st.sampled_from(["cells.csv",
                                                  "lexemes.csv",
                                                  "sounds.csv",
                                                  "graphemes.csv"
                                                  ]), unique=True))
    if "cells.csv" in extra_tables:
        files["cells.csv"] = pd.DataFrame(cells, columns=["cell_id"])
        files["cells.csv"].set_index("cell_id")
        m["resources"].append({"path": "cells.csv",
                               "name": "cells"})
    if "lexemes.csv" in extra_tables:
        files["lexemes.csv"] = draw(lexemes_table(lexemes, sounds))
        m["resources"].append({"path": "lexemes.csv",
                               "name": "lexemes"})
    if "sounds.csv" in extra_tables:
        files["sounds.csv"] = pd.DataFrame(sounds["C"] + sounds["V"],
                                           columns=["sound_id"])
        files["sounds.csv"].set_index("sounds_id")
        m["resources"].append({"path": "sounds.csv",
                               "name": "sounds"})
    if "graphemes.csv" in extra_tables:
        files["graphemes.csv"] = pd.DataFrame([Cs[c] for c in sounds["C"]] +
                                              [Vs[v] for v in sounds["V"]],
                                              columns=["grapheme_id"])
        files["graphemes.csv"].set_index("grapheme_id")
        m["resources"].append({"path": "graphemes.csv",
                               "name": "graphemes"})
    if draw(st.booleans()):
        fvs = pd.DataFrame([
            [v, f, i] for f in feats for i, v in enumerate(feats[f])
        ], columns=["value_id", "feature", "canonical_order"])
        fvs["label"] = fvs["value_id"] + "ive"
        fvs = fvs.set_index("value_id")
        files["feature_values.csv"] = fvs
        m["resources"].append({"path": "feature_values.csv",
                               "name": "features-values"})

    files["paralex.package.json"] = m
    return files


@st.composite
def forms_df(draw, sounds, lexemes, cells):
    """ Strategy to generate a synthetic forms dataframe. """
    max_affixes = max(1, len(lexemes) // 2)
    # sample affixes
    affixes = {c: draw(st.sets(affix(sounds), min_size=1, max_size=max_affixes)) for c in cells}

    def concatenate_words(a, is_suffix, b):
        phon, orth = zip(a, b) if is_suffix else zip(b, a)
        return " ".join(phon), "".join(orth)

    rows = []

    i = 0
    for lex in lexemes:
        for c in cells:
            exp = draw(st.lists(st.sampled_from(list(affixes[c])), min_size=1, max_size=3, unique=True))
            for a in exp:
                i += 1
                rows.append([
                    "f" + str(i),
                    lex[1],
                    c,
                    *concatenate_words(lex, *a)
                ])
    df = pd.DataFrame(rows, columns=["form_id", "lexeme", "cell", "phon_form", "orth_form"])
    return df


@st.composite
def metadata(draw):
    """ Strategy to generate fake metadata for a dataset

    This includes a list of resources with paths & names,

    """
    meta = {"resources": [],
            "title": draw(st.text(min_size=5, max_size=20)),
            "description": draw(st.text(min_size=10, max_size=50)),
            "profile": "data-package",
            }
    if draw(st.booleans()):
        meta["contributors"] = [
            {
                "title": draw(st.text(min_size=5, max_size=10)) + " " + draw(st.text(min_size=5, max_size=10)),
            } for _ in range(draw(st.integers(min_value=1, max_value=6)))
        ]
    if draw(st.booleans()):
        meta["keywords"] = draw(st.lists(st.text(min_size=3, max_size=10)))
    return meta



import pandas as pd
import frictionless as fl
from pathlib import Path
from slugify import slugify
import re


def create_ids(df, pattern="{}"):
    def to_id(row):
        id = "_".join(row)
        return slugify(pattern.format(id))

    ids = df.apply(to_id, axis=1)
    dups = ids.groupby(ids).rank(method="first", ascending=False).apply(lambda x: "" if x == 1 else str(int(x)))
    return ids + dups


def segment_sounds(col, sounds):
    def splitter(series, split_pattern):
        series = series.str.split(pat=split_pattern, regex=True)
        return series.apply(lambda x: " ".join([char for char in x if char]))

    defs = col == "#DEF#"
    split_pattern = re.compile("(" + "|".join(sorted(sounds, key=len, reverse=True)) + ")")
    segmented = col.copy()
    segmented[~defs] = splitter(segmented[~defs], split_pattern)
    return segmented


def read_table(name, package, **kwargs):
    """
    Reads Paralex tables.

    Parameters:
        name(str): A table's name.
        package: A frictionless Package.
        **kwargs: keyword arguments passed to `pandas.read_csv`
    Returns:
        a `pandas.DataFrame` containing the required table.
    """
    table_data = package.get_resource(name)
    paths = [table_data.path]
    if table_data.scheme == "multipart":
        paths.extend(table_data.extrapaths)
    return read_table_from_paths(paths, package, **kwargs)


def read_table_from_paths(paths, package, **kwargs):
    paths = paths if isinstance(paths, list) else [paths]
    return pd.concat(
        [pd.read_csv(absolute_path(package, path), **kwargs) for path in paths]
        , axis=0)

def absolute_path(package, path):
    return Path(package.basepath) / path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module provides tool to read and download existing paralex data.
"""

import io
import json
import glob
import tempfile
import zipfile
from pathlib import Path
import requests
import pandas as pd
import logging
from iso639 import Language, LanguageNotFoundError
from openscm_zenodo.zenodo import ZenodoInteractor
# Our modules
from .paths import docs_path

log = logging.getLogger()
unknown = "*unknown*"

zenodo_interactor = ZenodoInteractor()


def _get_glottocode():
    glottolog_langs = "https://raw.githubusercontent.com/glottolog/glottolog-cldf/refs/tags/v5.1/cldf/languages.csv"
    glottocode = pd.read_csv(io.BytesIO(requests.get(glottolog_langs).content))
    return glottocode


def _save_record_files(record, directory, expected_files=None):
    """
    Save files from a Zenodo record.

    Arguments:
        record (dict): A Zenodo record.
        directory (str): Path to the directory where the files should be saved.
        expected_files: Set of file extensions to download. If None, downloads all files.
    """
    directory.mkdir(exist_ok=True, parents=True)
    files = []
    log.info('Downloading files from the Zenodo record.')
    for file in requests.get(record["links"]["files"]).json()["entries"]:
        file_name = file["key"]
        file_link = file["links"]["content"]
        if file_name.endswith(".zip"):
            with zipfile.ZipFile(io.BytesIO(requests.get(file_link).content)) as zfile:
                zfile.extractall(directory)
                files.extend([f.filename for f in zfile.infolist()])
        elif (expected_files is None) or any(file_name.endswith(e) for e in expected_files):
            with (directory / file_name).open("wb") as f:
                f.write(requests.get(file_link).content)
                files.append(file_name)
    log.info('Files downloaded.')
    return files


def _get_zenodo_datasets():
    """
    Reads the list of available Paralex datasets from Zenodo.

    Returns:
        List[dict]: A list of JSON records.
    """
    log.info('Sending Zenodo request...')
    response = requests.get('https://zenodo.org/api/records',
                            params={'communities': 'paralex'})

    # TODO exit nicely in case of bad answer
    log.info('Received response from Zenodo')
    return response.json()["hits"]["hits"]


def _get_previous_data():
    """
    Tries to read the data from the last run.
    Exposed on the Paralex website.
    """

    ######### HACK TO GET LOCAL FILE
    # url = "http://localhost:8000/paralex/known-datasets.csv"
    url = "https://paralex-standard.org/known-datasets.csv"
    log.info('Fetching previous list of datasets.')
    try:
        df = pd.read_csv(url, index_col=0)
        return df
    except Exception as e:
        print(f"{e} occured while fetching {url}")
        return None


def _gather_dataset_info(glottolog_data, save=True, update=False):
    """
    Gets information on all datasets and stores in a DataFrame.

    Arguments:
        save (bool): Whether to save the dataframe as CSV
            or to return the object.
        update (bool): Whether to force update of the data.
    """
    old = _get_previous_data()
    records = _get_zenodo_datasets()
    data = []
    fields = ['id', 'permanent_id', 'name', 'lang', 'paralex_version', 'version', 'contributors', 'title', 'doi', 'comment']
    for record in records:
        rec_id = record['id']
        if (
                update
                or (old is None)
                or not (rec_id in old.index)
                or not ('permanent_id' in old.columns)):

            package_infos = _get_package_infos(record)

            ######### TEMPORARY HACK TO FAKE UPDATED DATASETS
            tmp_replacer = {"PrinParLat": ["lat"], "LatInfLexi": ["lat"], "LeFFI": ["ita"]}
            package_infos["lang"] = tmp_replacer.get(package_infos["name"], package_infos["lang"])
            ######### END TEMPORARY HACK TO FAKE UPDATED DATASETS

            data.append(package_infos)
        else:
            package_infos = old.loc[old.index == rec_id, :].reset_index().iloc[0].to_dict()
            package_infos = {key: value for key, value in package_infos.items() if key in fields}
            package_infos['lang'] = old.loc[old.index == rec_id, 'lang'].to_list()
            data.append(package_infos)
    data = pd.DataFrame(data).explode("lang").fillna(unknown).set_index('id')

    def get_language(lg):
        try:
            return Language.match(lg)
        except LanguageNotFoundError:
            return None

    log.info('Matching language names.')
    data["language_norm"] = data["lang"].apply(get_language)
    glottolog_data["language_norm"] = glottolog_data["ISO639P3code"].apply(get_language)
    res = data.join(glottolog_data.set_index("language_norm").loc[:, ["Latitude", "Longitude", "Family_ID"]], "language_norm",
                    how="left")
    res['lang_name'] = res.language_norm.apply(lambda lang: lang.name if lang else unknown)
    # Normalize names
    res.columns = [c.lower() for c in res.columns]
    # Get family name
    lang_labels = dict(glottolog_data.loc[:, ["ID", "Name"]].to_records(index=False))
    res["family"] = res["family_id"].map(lang_labels)
    log.info('Finnished gathering datasets.')
    if save:
        res.to_csv(docs_path / "known-datasets.csv")
    else:
        return res


def _get_package_file(record):
    def load_json(filepath):
        try:
            with filepath.open(encoding="utf-8") as fp:
                return json.load(fp)
        except Exception as e:
            print(f"{e} occured at {filepath}")
        return {}

    def is_package_file(json):
        return json.get("profile", None) == "data-package"

    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = Path(tempdir)
        _save_record_files(record, tempdir, expected_files={".csv", ".json", ".md", ".bib"})  # Save all files to temp dir, unpacking zips

        # Search for json files
        files = list(tempdir.glob("*package.json"))
        if len(files) == 0:
            files = list(tempdir.glob("**/*.json"))

        # Find the package
        for file in files:
            md = load_json(file)
            if is_package_file(md):
                return md
    return None


def _get_package_infos(record):
    rec_id = record['id']
    pid = record['conceptrecid']
    doi = record["links"]["doi"]
    title = record["title"]
    log.info(f"Updating data for package '{title}'.")
    json_data = _get_package_file(record)
    if json_data is not None:
        return {"title": title,
                "id": rec_id,
                'permanent_id': pid,
                "doi": doi,
                "name": json_data.get("name", json_data.get("title", unknown)),
                "lang": json_data.get("languages_iso639", unknown),
                "version": json_data.get('version', unknown),
                "paralex_version": json_data.get("paralex-version", unknown),
                "contributors": ";".join(c["title"] for c in json_data["contributors"]),
                }

    return {"title": title,
            "id": rec_id,
            'permanent_id': pid,
            "doi": doi,
            "comment": "no valid metadata found"
            }


def list_datasets(args):
    """
    Provides a list of all Paralex datasets that are available
    """
    glottocode_data = _get_glottocode()
    df = _gather_dataset_info(glottocode_data, save=False, update = args.update)
    if args.iso:
        log.info('Returning datasets for ISO code %s only', args.iso)
        df = df[df.lang.isin(args.iso)]
    df = df[['permanent_id', 'doi', 'lang', 'lang_name', 'family', 'version', 'paralex_version']]
    if args.output:
        df.to_csv(args.output)
    print(df.to_markdown())

    log.info('You can download a specific dataset with `paralex get <ZENODO_ID>`')


def download_dataset(args):
    """
    Downloads a dataset and extracts the zip archive if necessary.
    """

    log.info('Fetching a list of files to download')
    out = args.output if args.output else f"{args.zenodo_id}/"
    record = zenodo_interactor.get_record(args.zenodo_id).json()
    _save_record_files(record, Path(out))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Handles command line interface
"""
import argparse
from .gendoc import _make_standard_package, _write_doc
from .validate import paralex_validation
from .meta import gen_metadata
from .explorer import list_datasets, download_dataset
import logging

logging.basicConfig(level=logging.INFO)


def parse_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='Valid subcommands')

    # Subparser to build the standard
    make_std = subparsers.add_parser("make_standard",
                                     help='Builds the Paralex standard.')
    make_std.set_defaults(func=_make_standard_package)
    make_doc = subparsers.add_parser("make_doc")
    make_doc.set_defaults(func=_write_doc)

    # Subparser to validate a package
    validate = subparsers.add_parser("validate",
                                     help='Validates a Paralex package against a JSON file.')
    validate.add_argument("package", help="JSON description of the package to validate")
    validate.add_argument('--basepath', type=str,
                          help='Basepath for the package to validate.')
    validate.set_defaults(func=paralex_validation)

    # Subparser to generate metadata
    meta = subparsers.add_parser("meta",
                                 help='Generates a JSON descriptor from a YAML file.')
    meta.add_argument("config", help="YAML description of the package to describe.")
    meta.add_argument('--basepath', type=str,
                      help='Basepath for the generated .package.json file.')
    meta.set_defaults(func=gen_metadata)

    # Subparser to gather dataset info
    gather = subparsers.add_parser("list",
                                   help='Lists all available Paralex datasets.')
    gather.add_argument("--iso", action="extend", nargs="+", type=str, help="Language codes to catch.")
    gather.add_argument("-o", '--output', type=str, help="Output file. Saves the data to a CSV file.")
    gather.add_argument("-u", '--update', help="Force update of all the dataset metadata.",
                        action="store_true", default=False)
    gather.set_defaults(func=list_datasets)

    # Subparser to download a dataset
    get = subparsers.add_parser("get",
                                help='Downloads a Paralex dataset.')
    get.add_argument("zenodo_id", type=str, help="Zenodo ID of the dataset to download.")
    get.add_argument("-o", '--output', type=str, help="Output folder. Defaults to working directory")
    get.set_defaults(func=download_dataset)
    return parser.parse_args()


logo = r"""
 &&&&&&&&&&&&&&&&&
&                 &      /#######                              /##
&  ---  ---  ---  &     | ##__  ##                            | ##
&  ---  ---  ---  &     | ##  \ ## /######   /######  /###### | ##  /######  /##   /##
&  ---  ---  ---  &     | #######/|____  ## /##__  ##|____  ##| ## /##__  ##|  ## /##/
&                 &     | ##____/  /#######| ##  \__/ /#######| ##| ######## \  ####/
 &&&&&&&&&&&&&&   &     | ##      /##__  ##| ##      /##__  ##| ##| ##_____/  >##  ##
               &  &     | ##     |  #######| ##     |  #######| ##|  ####### /##/\  ##
                 &&     |__/      \_______/|__/      \_______/|__/ \_______/|__/  \__/

"""


def main():
    args = parse_args()
    if len(args.__dict__) == 0:
        print(logo + 'Please type "paralex --help" to display available commands.')
    else:
        args.func(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module manipulates frictionless metadata.
"""
import collections.abc
import inspect
import json
import yaml
import logging
import re
from collections import defaultdict
from pathlib import Path
import frictionless as fl

from .utils import read_table_from_paths
from . import standard_path, VERSION

relations = ["isCitedBy", "cites", "isSupplementTo", "isSupplementedBy", "isContinuedBy", "continues", "isDescribedBy",
             "describes", "hasMetadata", "isMetadataFor", "isNewVersionOf", "isPreviousVersionOf", "isPartOf",
             "hasPart", "isReferencedBy", "references", "isDocumentedBy", "documents", "isCompiledBy", "compiles",
             "isVariantFormOf", "isOriginalFormof", "isIdenticalTo", "isAlternateIdentifier", "isReviewedBy", "reviews",
             "isDerivedFrom", "isSourceOf", "requires", "isRequiredBy", "isObsoletedBy", "obsoletes"]
resource_name_pattern = re.compile(r'^([-a-z0-9._/])+$')


def gen_metadata(args):
    with open(args.config, 'r') as infile:
        config = yaml.safe_load(infile)

    if args.basepath:
        basepath = Path(args.basepath)
    elif "basepath" in config.keys():
        # Keeping this for backward compatibility.
        logging.warning("Specifying the basepath in the config file is deprecated. Use the --basepath argument instead.")
        basepath = Path(config['basepath'])
    else:
        logging.warning("No basepath was passed. Using the YAML config file location.")
        basepath = Path(args.config).parent
    config['basepath'] = str(basepath.resolve())
    path = basepath.resolve() / f"{config['name']}.package.json"

    package = paralex_factory(**config)
    logging.info('Writing the package definition to %s', path)
    package.to_json(path)


def _to_pattern(iterable, sep="", diacritics=None):
    iterable = sorted(iterable, key=len, reverse=True)
    items = "|".join(iterable)
    if diacritics is None or not diacritics:
        char = f"({items})"
    else:
        diacs = "|".join(diacritics)
        char = f"(({diacs})*({items})({diacs})*)"

    if sep == "":
        return f"{char}+"
    else:
        return f"{char}({sep}{char})*"


def _get_default_package(basepath=""):
    """ Instantiate a default paralex frictionless package

    This uses the paralex json file, removing specific instance information
    (version, name, contributors, related_identifiers).

    Returns (fl.Package): a default package for paralex lexicons

    """
    default_to_remove = ["version", "name", "contributors"]
    default = json.load((standard_path / "paralex.package.json").open("r", encoding="utf-8"))

    # Remove default dummy values in package
    for key in default_to_remove:
        del default[key]
    return fl.Package(default, basepath=basepath)


def default_update(specific, default):
    """ Update a default package with specific information.

    The behaviour is as follows:
        - For the "Fields" list, append any extra column from default
         which are not already specified in the specific information
        - Otherwise for all mappings: recursively update
        - Otherwise add information from default if it does not exist in specific.

    Args:
        specific (dict): specific information (user specified)
        default (dict): default information (from paralex standard)

    Returns:
        a dictionary of the default information, updated with the specific information.

    """
    for k, v in default.items():
        if isinstance(v, list) and k == "fields":
            cols = specific.get(k, [])
            specific[k] = cols + list(filter(lambda c: c not in cols, v))
        if isinstance(v, collections.abc.Mapping):
            specific[k] = default_update(specific.get(k, {}), v)
        elif specific.get(k, None) is None:
            specific[k] = v
    return specific


def package_from_kwargs(basepath=False, **kwargs):
    """
    Arguments:
        basepath (str): Absolute path where the metadata file will be stored.

    Returns:
        frictionless package
    """
    if not basepath:
        logging.warning('No basepath was passed, assuming current location.')
        basepath = str(Path().resolve())

    # Separate kwargs between pre-defined ones and custom ones
    defined = {x.name: kwargs[x.name] for x in fl.Package.__attrs_attrs__ if x.name in kwargs}
    custom = {k: kwargs[k] for k in kwargs if k not in defined}
    p = fl.Package(**defined, basepath=basepath)
    p.custom.update(custom)
    return p


def paralex_factory(title, files=None, tables=None, **kwargs):
    """ Generate frictionless metadata for this paralex lexicon.

    Use the default package, fillin-in user specified information.
        Keeps only existing tables,
        updates foreign keys and table relations according to which tables are
        declared.

    Args:
        title (str): a title for the package
        files (dict): dictionary of file names to resource information
        tables (dict): deprecated, see `files`
        **kwargs:

    Returns:

    """
    ## Update package infos
    kwargs["title"] = title
    if files is None:
        if tables is None:
            raise ValueError("Paralex package must have some files (tables, etc).")
        files = tables
        logging.warning("Deprecated ! As of Paralex v.2.0.8,"
                        " the `tables` keyword is deprecated in config files, "
                        "and replaced by `files`.")

    # Get a frictionless package instantiated from the standard json
    default_package = _get_default_package()
    default_package_infos = default_package.to_dict()
    del default_package_infos["resources"]
    default_update(kwargs, default_package_infos)
    p = package_from_kwargs(**kwargs)

    ## Update resources infos
    default_resources = set(default_package.resource_names)
    resources = set(files)
    added_resources = resources - default_resources
    missing_resources = default_resources - resources

    for res in resources:
        if resource_name_pattern.match(res) is None:
            suggested = re.sub("[^-a-z0-9._/]", "", res.lower())
            raise ValueError(f"Table name {repr(res)} is invalid. "
                             "\nTable names can only contain '-', '_', '.', "
                             "numbers or lowercase latin letters. "
                             f"Try instead something like:\n\tparalex_factory({title}, "
                             "{"
                             f" ... '{suggested}': {files[res]} ... "
                             "} ... )")
        res_metadata = files[res]
        res_metadata["name"] = res

        # Grabbing just the first path, assuming resources are homogeneous !
        path = res_metadata["path"]
        path = path[0] if isinstance(path, list) else path

        if res in added_resources:
            logging.warning(f"Adding a table which is not "
                            f"specified in the standard: {repr(res)}.\n\t"
                            f"Did you mean one of: `{'`, `'.join(missing_resources)}`?")
        else:
            default_table = default_package.get_resource(res)
            default_update(res_metadata, default_table.to_dict())

        if path.endswith(".csv"):
            table_contents = read_table_from_paths(res_metadata["path"], p)
            cols = list(table_contents.columns)

            def foreign_key_exists(key_spec):
                table_ref = key_spec["reference"]["resource"]
                col_names = key_spec["fields"]
                return table_ref in resources and all(c in cols for c in col_names)

            ## Adjust foreign keys according to existing tables
            fkeys = list(filter(foreign_key_exists,
                                res_metadata.get("schema", {}).get("foreignKeys", [])))
            if fkeys:
                res_metadata["schema"]["foreignKeys"] = fkeys
            elif "schema" in res_metadata and "foreignKeys" in res_metadata["schema"]:
                del res_metadata["schema"]["foreignKeys"]

            # Remove standard columns which are not present
            new_table_fields = list(res_metadata["schema"]["fields"])
            for field in res_metadata["schema"]["fields"]:
                if field["name"] not in cols:
                    new_table_fields.remove(field)
            res_metadata["schema"]["fields"] = new_table_fields

        r = fl.Resource(res_metadata, detector=fl.Detector(schema_sync=True))

        p.add_resource(r)

    ## Set Foreign keys for tags

    taggable_tables = {"forms", "lexeme", "frequencies"} & resources

    for tab_name in taggable_tables:
        res = p.get_resource(tab_name)
        tag_cols = {col for col in res.schema.field_names
                    if col.endswith("_tag")}
        if tag_cols:
            if "tags" not in resources:
                logging.warning(f"You have tag columns ({', '.join(tag_cols)}) "
                                f"in table {tab_name},"
                                f"but no tags table.")
            else:
                by_tag = defaultdict(list)
                for r in p.get_resource("tags").read_rows():
                    by_tag[r["tag_column_name"]].append(r["tag_id"])
                for col in tag_cols:
                    if col not in by_tag:
                        logging.warning(f"No tags defined for column {col} in table {tab_name}. "
                                        f"Please edit the tags table.")
                    else:
                        col_def = res.schema.get_field(col)
                        constraints = col_def.constraints
                        constraints["pattern"] = _to_pattern(by_tag[col], sep="\\|")
                        col_def.constraints = constraints

    forms_table = p.get_resource("forms")

    # Set pattern for orth_form
    if {"graphemes", "forms"} <= resources \
            and forms_table.schema.has_field("orth_form"):
        graphemes = [r["grapheme_id"] for r in p.get_resource("graphemes").read_rows()]
        form = forms_table.schema.get_field("orth_form")
        constraints = getattr(form, "constraints")
        constraints["pattern"] = _to_pattern(graphemes, sep="")
        form.constraints = constraints

    # Set pattern for phon_form
    if {"sounds", "forms"} <= resources \
            and forms_table.schema.has_field("phon_form"):
        seg = []
        supra_seg = []
        for row in p.get_resource("sounds").read_rows():
            s = row["sound_id"]
            if s.startswith("#") or s.endswith("#"):
                continue
            if row.get("tier", "segmental") == "segmental":
                seg.append(s)
            else:
                supra_seg.append(s)
        form = forms_table.schema.get_field("phon_form")
        constraints = form.constraints
        constraints["pattern"] = _to_pattern(seg, diacritics=supra_seg, sep=" ")

    # Set pattern for cells identifiers (combinations of features)
    if {"cells", "features-values"} <= resources:
        features = [r["value_id"] for r in p.get_resource("features-values").read_rows()]
        cell_id = p.get_resource("cells").schema.get_field("cell_id")
        constraints = cell_id.constraints
        constraints["pattern"] = _to_pattern(features, sep=r"\.")

    p.infer()
    p.custom["paralex-version"] = VERSION
    return p
